import re
import time

from cgi import escape as escent

from UserDict import UserDict

from twisted.internet.defer import Deferred

from twisted.web2.stream import IByteStream

from zope.interface import implements

from xmlreader import Item, Dictable, Listable

import urllib

bardiv = '<div class="progress" style="text-align:center; width: %f%%;">%.2f%%</div>'

cmd = """
        <kukit:commands>
            <kukit:command selector="#%s_progressbar" name="replaceInnerHTML" selectorType="">
                <kukit:param name="html">%s</kukit:param>
            </kukit:command>
            <kukit:command selector="#%s_speed" name="replaceInnerHTML" selectorType="">
                <kukit:param name="html">%.2fKB/s</kukit:param>
            </kukit:command>
        </kukit:commands>
        <script>
            kukit.feeder.feed();
        </script>
"""


class FileUploadConfig(Item):
    def _leave(self):
        print "ProxyConfig._leave"
        proxyconf = self._aq('proxy')
        feederconf = self._aq('feeder')
        hw = self._getRoot()
        urlRe= re.compile(self.urlRe)
        fu = FileUpload(urlRe, feederconf.node)
        
        proxyconf.addRequestHook(('POST',), fu.beforeHook)

classMap = dict(
    fileupload = FileUploadConfig,
)

class MimeDecoder(UserDict):
    def __init__(self):
        UserDict.__init__(self)
        self.state = 'init'
        self.prevdata = ""
        self.inputInfo = {}
    
    def _doInit(self, data):
        self.delim, data = data.split('\r\n', 1)
        self.state = 'headers'
        self.curheaders = {}
        return data
    
    def _doNext(self, data):
        line, data = data.split('\r\n', 1)
        if not line.startswith(self.delim):
            raise Exception("Bug in the state machine")
        if line == self.delim+'--':
            self.state='finished'
        else:
            self.curheaders = {}
            self.state='headers'
        return data
    
    def _doHeaders(self, data):
        line, data = data.split('\r\n', 1)
        if len(line)>0:
            key, val = line.split(': ')
            if "; " in val:
                val, addit = val.split('; ', 1)
                addit = [item.rstrip('"').split('="') for item in addit.split('; ')]
                self.curheaders.update(addit)
            self.curheaders[key] = val
        else:
            self.inputInfo[self.curheaders['name']] = self.curheaders
            if 'filename' in self.curheaders:
                self.state = 'file'
            else:
                self.state = 'simpleinput'
        return data
    
    def _doFile(self, data):
        if ('\r\n'+self.delim) in data:
            chunk, data = data.split('\r\n'+self.delim, 1)
            data = self.delim + data
            self.state = 'next'
        else:
            return ""
    
    def _doSimpleinput(self, data):
        line, data = data.split('\r\n', 1)
        self[self.curheaders['name']]=line
        self.state = 'next'
        return data
    
    def _doFinished(self, data):
        if data=='\r\n':
            return ""
        else:
            raise Exception("Bug in the state machine")
    
    def feed(self, data):
        data = self.prevdata + data
        while ('\r\n' in data) or ((self.state=='data') and (len(data)>0)):
            hdlr = getattr(self, '_do'+self.state.title())
            data = hdlr(data)
        self.prevdata=data

class FileUploadStream:
    implements(IByteStream)
    
    def __init__(self, original, feeder):
        self.original = original
        self.length = original.length
        self.mimedecode = MimeDecoder()
        self.readlen = 0
        self.lasttime = None
        self.feeder = feeder
    
    def read(self):
        res = self.original.read()
        if isinstance(res, Deferred):
            defer = Deferred()
            res.addCallback(self.realread, defer)
            res = defer
        else:
            res = self.realread(res)
        return res
    
    def realread(self, data, defer=None):
        if data is None:
            return data
        self.readlen += len(data)
        if 'eventPushUid' in self.mimedecode:
            self.notify(len(data))
        else:
            self.mimedecode.feed(data)
        if not defer is None:
            defer.callback(data)
        else:
            return data
    
    def notify(self, length):
        now = time.time()
        if self.lasttime is None:
            self.lasttime = now
            self.firsttime = now
            return
        if (int(now) > int(self.lasttime)) or (self.readlen == self.length):
            perc = self.readlen*100.0/self.length
            speed = self.readlen/1000.0/(now-self.firsttime)
            print "%.2f%% at %.2fKB/s" % (perc,
                                          speed)
            data = cmd % (self.mimedecode['fieldName'],
                          escent(bardiv % (perc, perc)),
                          self.mimedecode['fieldName'],
                          speed)
            self.feeder.pushUp(self.mimedecode['eventPushUid'],
                               data)
        self.lasttime = now
    
    def split(self, point):
        res = self.original.split(point)
        return res
    
    def close(self):
        res = self.original.close()
        return res

class FileUpload:
    
    def __init__(self, urlRe, feeder):
        self.urlRe = urlRe
        self.feeder = feeder
    
    def beforeHook(self, meth, url, hdrs, data):
        if not self.urlRe.search(urllib.unquote(url)):
            return (meth, url, hdrs, data)
        return (meth, url, hdrs, FileUploadStream(data, self.feeder))
