
from twisted.internet import reactor, protocol

from twisted.internet.defer import Deferred

from twisted.web2 import resource
from twisted.web2 import responsecode
from twisted.web2.client import http as client2
from twisted.python import log

from xmlreader import Item, Dictable, Listable

import urllib


class BackEndConfig(Dictable):
    def _enter(self):
        print "BACKEND", self.nickname, "enter"
    def _leave(self):
        hw = self._getRoot()
        backendmgr = BackendManager()
        print "BACKEND", self.nickname, "leave"
        hw.register("IBackEnd", self.nickname, backendmgr)
        for host in self.host:
            backendmgr.registerHost(host.address, int(host.port))
        backendmgr.maxconns=getattr(self, "maxconns", 10)

class ProxyConfig(Item):
    path='/'
    def _leave(self):
        hw = self._getRoot()
        print "ProxyConfig._leave"
        backend = hw.getRegistered('IBackEnd', self.backend)
        self.vsite = ProxyNode(backend, str(self.rewriteURI))
        self._aq('vhostobj').registerWebSite(self.path, self.vsite)
    
    def addRequestHook(self, methods, hook):
        self.vsite.addRequestHook(methods, hook)

class Zope2RewriteConf(Item):
    prefix = "/VirtualHostBase/http/%(frontHost)s:%(frontPort)d"
    suffix = "/VirtualHostRoot"
    
    def _leave(self):
        self.__parent__.rewriteURI = self.prefix.rstrip('/') + self.pathToSite.rstrip('/') + self.suffix
        print self._aq('rewriteURI')

classMap = dict(
    backend = BackEndConfig,
    proxy = ProxyConfig,
    host = Listable,
    zope2VHRewrite = Zope2RewriteConf,
)


class BackendManager(object):
    def __init__(self):
        self.hostconns = {}
        self.busyconns = []
        self.idleconns = []
        self.hostlist = []
        print "BackendManager __init__(%s)" % repr(self)
    
    def clientBusy(self, proto):
        #print "BackendConn.clientBusy", proto
        #self.idleconns.remove(proto)
        #self.busyconns.append(proto)
        pass
    
    def clientIdle(self, proto):
        #print "BackendConn.clientIdle", proto
        #self.busyconns.remove(proto)
        self.idleconns.append(proto)

    def clientPipelining(self, proto):
        #print "BackendConn.clientPipelining", proto
        pass

    def clientGone(self, proto):
        #print "BackendConn.clientGone", proto
        #try:
        #    self.busyconns.remove(proto)
        #except:
        #    pass
        try:
            self.idleconns.remove(proto)
        except:
            print "Not idle", proto
    
    def registerHost(self, host, port):
        self.hostconns[(host,port)]=[]
        self.hostlist.append((host,port))
    
    def getConn(self):
        if  self.idleconns:
            return self.idleconns.pop(0)
        
        #print "MAKING A NEW CONNECTION"
        address, port = None, None
        for key, conns in self.hostconns.items():
            if len(conns)>self.maxconns:
                continue
            address, port = key
        
        if not address:
            raise Exception
        
        print "CONNECTING TO", address, port
        creator = protocol.ClientCreator(reactor,
                                         client2.HTTPClientProtocol,
                                         manager = self)
        return creator.connectTCP(address, port)
    
    def submitRequest(self, method, url, headers, stream):
        myDefer = Deferred()
        def sendReqs(proto):
            print "SENDING A REQUEST", method, url
            def gotResp(resp):
                #print "GOT A RESPONSE"
                myDefer.callback(resp)
            proto.submitRequest(client2.ClientRequest(
                method,
                url,
                headers,
                stream
            ), closeAfter=False).addCallbacks(
                callback=gotResp,
                errback=lambda error: myDefer.errback(error),
            )
        try:
            proto = self.getConn()
        except Exception, e:
            print e.__class__.__name__, e
        if isinstance(proto, Deferred):
            #print "CLIENT DEFER", proto
            proto.addCallbacks(callback=sendReqs,
                               errback=lambda error: myDefer.errback(error))
        else:
            #print "sendReqs"
            sendReqs(proto)
        return myDefer

class ProxyNode(resource.Resource):
    
    def __init__(self, backend, rewriteURI):
        resource.Resource.__init__(self)
        self.backend = backend
        self.urlRewriteFmt = rewriteURI
        self.requesthooks={}
    
    def addRequestHook(self, methods, hook):
        for meth in methods:
            if not meth in self.requesthooks:
                self.requesthooks[meth] = []
            self.requesthooks[meth].append(hook)
    
    def locateChild(self, request, stack):
        request.proxyPath = urllib.quote('/'.join(stack))
        return self, ()
    
    def getRewrParams(self, request):
        return dict(
            frontHost = request.host,
            frontPort = request.port,
        )

    def makeUrl(self, request):
        url = urllib.quote(self.urlRewriteFmt % self.getRewrParams(request))
        url= '/'.join((url, request.proxyPath))
        origurl, params = urllib.splitquery(request.uri)
        if params:
            url = "?".join((url, params))
        return url
    
    def http_POST(self, request):
        print request.proxyPath
        fwdHeaders=request.headers
        args = (
            "POST",
            self.makeUrl(request),
            fwdHeaders,
            request.stream)
        for hook in self.requesthooks.get('POST', ()):
            args = hook(*args)
        return self.backend.submitRequest(*args)

    def http_GET(self, request):
        if request.stream.length != 0:
            return responsecode.REQUEST_ENTITY_TOO_LARGE
        fwdHeaders=request.headers
        args = (
            "GET",
            self.makeUrl(request),
            fwdHeaders,
            None)
        for hook in self.requesthooks.get('GET', ()):
            args = hook(*args)
        return self.backend.submitRequest(*args)
