
"""
VirtualHosts
"""

import time
import re

from twisted.web2 import http, resource
from twisted.web2 import http_headers, responsecode

from twisted.web2 import server

from xmlreader import Listable, PlacefulSingleton, Class, Dictable


class VirtualHostConfig(Dictable):
    def _enter(self):
        self.vhostobj = self._aq('vhostobj').build()
        self.vhostobj.nickname = self.nickname
        
    def _leave(self):
        print "VirtualHostConfig._leave"
        hw = self._getRoot()
        for mount in self.mount:
            vhostmgr = hw.getRegistered("IFrontEnd", mount.frontend)
            vhostmgr.registerVirtualHost(self.vhostobj, mount)


classMap = dict(
    vhostmgr = Class,
    vhostobj = Class,
    mount = Listable,
    virtualhost = VirtualHostConfig,
    vhostname = Listable,
)

class RegexVHostMgr(resource.Resource):
    """
    The toplevel node
    """
    addSlash = True
    def render(self, request):
        """
        Renders the current node
        """
        #return http.Response(stream=self.content)
        content = "<html><body><center><h1>No VHost for http://%s:%d%s</h1></center></body></html>" % (
            request.host,
            request.port,
            request.uri)
        contenttype = http_headers.MimeType('text', 'html', charset= "utf-8")
        return http.Response(responsecode.OK,
            {'last-modified': time.time(),
            'etag': http_headers.ETag(str(hash(content))),
            'content-type': contenttype,},
            content)

    def locateChild(self, request, stack):
        for regex, vhost in self.vhostlist:
            #if request.port == port and regex.match(request.host):
            if regex.match(request.host):
                print "VIRTUAL HOST :", vhost.nickname
                return vhost, stack
        return self, ()

    def __init__(self):
        resource.Resource.__init__(self)
        self.vhostlist = []
        self.vhostmap = {}
    
    def buildSite(self):
        return server.Site(self)
    
    def registerVirtualHost(self, vhost, mount):
        if mount.regex in self.vhostmap.keys():
            print "Already registered VHost!!!"
            raise SystemExit
        self.vhostlist.append((re.compile(mount.regex), vhost))
        self.vhostmap[mount.regex] = vhost


class RPathNode(dict):
    """
    Vhost tree node
    """
    def __init__(self, website=False):
        dict.__init__(self)
        self.website = website

class RPathVHost(resource.Resource):
    """
    Vhost resource
    """
    addSlash = True

    def __init__(self):
        resource.Resource.__init__(self)
        self.tree = RPathNode()
    
    def registerWebSite(self, path, website):
        """
        call it to register a an IWebViewable object
        """
        path = path.rstrip('/').split('/')[1:]
        node = self.tree
        for name in path:
            if not name in node:
                node[name] = RPathNode()
            node = node[name]
        node.website=website
    
    def locateChild(self, request, stack):
        """
        part of IWebViewable interface
        """
        nodestack=[self.tree]
        for item in stack:
            if not item in nodestack[-1]:
                break
            nodestack.append(nodestack[-1][item])
        while nodestack:
            node = nodestack.pop()
            if node.website:
                #print "WEBSITE", "/".join([""]+stack), node.website.__class__.__name__
                return node.website, stack[len(nodestack):]
        return None, ()
