# -*- coding: utf-8 -*-

from xml.sax import ContentHandler

from xml.sax import make_parser
from xml.sax.handler import feature_namespaces


class XmlConfError(Exception): pass

class Item(object):
    isListable = False
    isDictable = False
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
    def _enter(self):
        pass
    def _leave(self):
        pass
    def _getRoot(self):
        node = self
        while node.__parent__:
            node = node.__parent__
        return node
    def _aq(self, name):
        node = self
        while not hasattr(node, name) and node.__parent__:
            node = node.__parent__
        return getattr(node, name)

class Dictable(Item):
    isDictable = True

class Listable(Item):
    isListable = True

class Module(Item):
    def __init__(self, modPath = None):
        #print "newClass", classPath
        if modPath is None:
            self.mod = None
            return
        try:
            self.mod = __import__(modPath, None, None, ["Eggs"])
        except ImportError:
            print "module", modPath, "not found"
            self.mod = None
            return
        if hasattr(self.mod, 'classMap'):
            print "updating tags", getattr(self.mod, 'classMap').keys()
            updateClassMap(getattr(self.mod, 'classMap'))

class Class(Item):
    def __init__(self, classPath = None):
        #print "newClass", classPath
        if classPath is None:
            self.build = None
            return
        classPath, className = classPath.rsplit(".", 1)
        classMod = __import__(classPath, None, None, ["Eggs"])
        try:
            self.build = getattr(classMod, className)
        except AttributeError:
            print "class", className, "not found in module", classMod
            print "\t"+"\n\t".join(dir(classMod))
            self.build = None

class DictableClass(Class):
    isDictable = True

class ListableClass(Class):
    isListable = True

class PlacefulSingleton(Class):
    def get(self):
        if not hasattr(self, "__singlinstance__"):
            self.__singlinstance__ = self.build()
        return self.__singlinstance__


classMap = dict(
    module = Module
)

class Finder(ContentHandler):
    def __init__(self, root):
        self.root     = root
        self.node     = root
        self.level    = 0
        self.isValue  = False
        self.isRoot   = True
   
    def printOut(self, line):
        print "  "*(self.level+2) + line
     
    def startElement(self, name, attrs):
        #print "  "*self.level, ">", name
        nodename = attrs.get('nickname', name)
        if self.isRoot:
            self.isRoot=False
            return
        if (len(attrs._attrs)==1) and attrs.has_key('value'):
            setattr(self.node, nodename, attrs._attrs['value'])
            self.isValue=True
        else:
            #print "Creating a", name,"with", attrs._attrs
            attrDict = dict([(str(key), val) for key,val in attrs.items()])
            #print attrDict
            node=classMap.get(name, Item)(**attrDict)
            #print node.isListable, node.isDictable
            if node.isListable:
                if not hasattr(self.node, name):
                    setattr(self.node,name, [])
                getattr(self.node, name).append(node)
            elif node.isDictable:
                if not hasattr(self.node, name):
                    setattr(self.node, name, {})
                #elif nodename in getattr(self.node, name):
                #    print nodename, "already in", getattr(self.node, name)
                #    raise XmlConfError
                getattr(self.node, name)[nodename] = node
            else:
                #if hasattr(self.node,nodename):
                #    print nodename, "already in", self.node
                #    raise XmlConfError
                setattr(self.node,nodename, node)
            #print "CREATED NODE", node, "of", self.node
            node.__parent__ = self.node
            node.__tagname__ = name
            node._enter()
            self.node = node
        self.level+=1
    
    def endElement(self, name):
        if self.isValue:
            self.isValue = False
        elif self.level != 0:
            self.node._leave()
            self.node = self.node.__parent__
        self.level -= 1
        #print "  "*self.level, "<", name

def parseXml(file, root = None):
    
    if root == None:
        root=Item()
        doReturn=True
    else:
        doReturn=False
    
    parser = make_parser()
    parser.setFeature(feature_namespaces, 0)
    dh = Finder(root)
    parser.setContentHandler(dh)
    parser.parse(file)
    
    if doReturn:
        return root

def updateClassMap(newmap):
    for tagname in set(classMap.keys()) & set(newmap.keys()):
        print "Overriding tagname :", tagname
    classMap.update(newmap)
